//=======================================================================
// Author: Donovan Parks
//
// Copyright 2010 Donovan Parks
//
// This file is part of NaiveBayes.
//
// NaiveBayes is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// NaiveBayes is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with NaiveBayes.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "stdafx.h"

#include "FastaIO.hpp"
#include "KmerCalculator.hpp"
#include "KmerModel.hpp"
#include "TaxonomyIO.hpp"
#include "Utils.hpp"

struct TopModel
{
	TopModel(): modelNum(0), logLikelihood(-1e10) {} 
	TopModel(uint _modelNum, float _logLikelihood): modelNum(_modelNum), logLikelihood(_logLikelihood) {}

	uint modelNum;
	float logLikelihood;
};

int main(int argc, char* argv[])
{
	if(argc != 6 && argc != 7)
	{
		std::cout << "nb-classify v1.0 by Donovan Parks, Norm MacDonald, and Rob Beiko" << std::endl;
		std::cout << std::endl;
		std::cout << "Usage: nb-classify <N> <T> <query fragment file> <model file> <results file>" << std::endl;
		std::cout << std::endl;
		std::cout << "N: is the n-mer length used while training models." << std::endl;
		std::cout << std::endl;
		std::cout << "T: the log likelihood of the top T models will be returned. If you wish to" << std::endl;  
		std::cout << "   have the log likelihood of all models in the results file set T = -1."<< std::endl; 
		std::cout << std::endl;
		std::cout << "query fragment file: multi-FASTA file containing query fragments to classify." << std::endl; 
		std::cout << std::endl;
		std::cout << "model file: text file indicating which models to use for classification." << std::endl;  
		std::cout << std::endl;
		std::cout << "results file: file where results will be written." << std::endl;   
		std::cout << std::endl;
		std::cout << std::endl;
		std::cout << "Typical usage:" << std::endl;
		std::cout << "  nb-classify 10 1 test.fasta ./models/models.txt results.txt" << std::endl;
		std::cout << std::endl;

		return 0;
	}

	int KMER_SIZE = atoi(argv[1]);
	int TOP_MODELS = atoi(argv[2]);
	std::string queryFragmentFile = argv[3];
	std::string inputModelFile = argv[4];
	std::string classificationFile = argv[5];
	
	bool bVerbose = false;
	if(argc == 7 && (std::string(argv[6]) == "VERBOSE" || std::string(argv[6]) == "verbose"))
		bVerbose = true;

	bool bRecordAllModels = false;
	if(TOP_MODELS == -1)
	{
		bRecordAllModels = true;
		TOP_MODELS = 1;
	}
	
	// classify query sequences
	std::cout << "Reading query sequences..." << std::endl;
	char* buffer = NULL;
	std::vector<SeqInfo> querySeqs;
	FastaIO fastaIO;
	fastaIO.readSeqs(queryFragmentFile, querySeqs, buffer, bVerbose);

	std::cout << "Calculating n-mers in query sequences..." << std::endl;
	KmerCalculator kmerCalculator(KMER_SIZE);
	std::vector< std::vector<uint> > queryKmerProfiles;
	queryKmerProfiles.reserve(querySeqs.size());
	for(uint seqIndex = 0; seqIndex < querySeqs.size(); ++seqIndex)
	{
		if(bVerbose)
			std::cout << querySeqs.at(seqIndex).seqId << std::endl;
		else if (seqIndex % 10000 == 0)
			std::cout << ".";

		std::vector<uint> profile;
		kmerCalculator.extractForwardKmers(querySeqs.at(seqIndex), profile);
		queryKmerProfiles.push_back(profile);
	}

	// apply each model to each query sequence
	std::cout << std::endl << std::endl;
	std::cout << "Applying models to query sequences:" << std::endl;
	std::ifstream modelStream(inputModelFile.c_str(), std::ios::in);

	uint modelNum = 0;
	std::vector< std::list<TopModel> > topModelsPerFragment(querySeqs.size());

	std::vector<std::string> modelNames;
	std::vector< std::vector<float> > modelLogLikelihoods;
	while(!modelStream.eof())
	{
		std::string line;
		std::getline(modelStream, line);

		if(line.empty())
			break;

		if(modelNum % 100 == 0)
		{
			std::cout.precision(3);
			std::cout << " " << modelNum << std::flush;
		}
		
		KmerModel kmerModel(KMER_SIZE);
		kmerModel.read(line);

		modelNames.push_back(kmerModel.name());

		ulong size = 0;
		if(bRecordAllModels)
			size = querySeqs.size();		
		std::vector<float> logLikelihoods(size);

		for(uint seqIndex = 0; seqIndex < querySeqs.size(); ++seqIndex)
		{
			// read sequence
			SeqInfo querySeqInfo = querySeqs.at(seqIndex);	

			// NB classifier
			float logLikelihood = kmerModel.classify(querySeqInfo, queryKmerProfiles[seqIndex]);

			// record models with highest log likelihood
			std::list<TopModel> topModels = topModelsPerFragment.at(seqIndex);

			if(topModels.size() == 0)
				topModels.push_front(TopModel(modelNum, logLikelihood));

			std::list<TopModel>::iterator it;
			for(it = topModels.begin(); it != topModels.end(); it++)
			{
				if(logLikelihood > it->logLikelihood)
				{
					topModels.insert(it, TopModel(modelNum, logLikelihood));
					break;
				}
			}

			if((int)topModels.size() > TOP_MODELS)
					topModels.pop_back();

			topModelsPerFragment.at(seqIndex) = topModels;

			if(bRecordAllModels)
				logLikelihoods[seqIndex] = logLikelihood;
		}

		if(bRecordAllModels)
			modelLogLikelihoods.push_back(logLikelihoods);	
	
		modelNum++;
	}
	std::cout << std::endl;

	// write out classification
	std::ofstream fout(classificationFile.c_str(), std::ios::out);	

	// check if all model results are to be written out
	if(bRecordAllModels)
	{		
		fout << "Fragment Id" << "\t" << "Length" << "\t" << "Valid n-mers";
		for(uint modelIndex = 0; modelIndex < modelNames.size(); ++modelIndex)
			fout << "\t" << modelNames[modelIndex];
		fout << std::endl;

		for(uint seqIndex = 0; seqIndex < querySeqs.size(); ++seqIndex)
		{
			SeqInfo querySeqInfo = querySeqs.at(seqIndex);

			fout << querySeqInfo.seqId << "\t" << querySeqInfo.length << "\t" << querySeqInfo.validKmers;

			for(uint modelIndex = 0; modelIndex < modelNames.size(); ++modelIndex)
				fout << "\t" << modelLogLikelihoods[modelIndex][seqIndex];
			fout << std::endl;
		}
	}
	else
	{
		for(uint seqIndex = 0; seqIndex < querySeqs.size(); ++seqIndex)
		{
			SeqInfo querySeqInfo = querySeqs.at(seqIndex);

			fout << querySeqInfo.seqId << "\t" << querySeqInfo.length << "\t" << querySeqInfo.validKmers;

			std::list<TopModel>::iterator it;
			for(it = topModelsPerFragment.at(seqIndex).begin(); it != topModelsPerFragment.at(seqIndex).end(); it++)
				fout << "\t" << modelNames[it->modelNum] << "\t" << it->logLikelihood;
		
			fout << std::endl;
		}
	}

	fout.close();

	delete[] buffer;

	return 0;
}

